package core;

import movement.MovementModel;
import movement.Path;
import routing.MessageRouter;
import routing.util.*;
import util.Ided;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import static core.Constants.DEBUG;
public class DTNHost extends Ided implements Comparable<DTNHost> {
    private static int nextAddress = 0;
    private int address;
    private Coord location;
    private Coord destination;
    private MessageRouter router;
    private MovementModel movement;
    private Path path;
    private double speed;
    private double nextTimeToMove;
    private String name;
    private List<MessageListener> msgListeners;
    private List<MovementListener> movListeners;
    private List<NetworkInterface> net;
    private List<Connection> allConnections;
    private ModuleCommunicationBus comBus;    private DTNFileSystem fileSystem;
    private boolean hasFileCapability;

    static {
        DTNSim.registerForReset(DTNHost.class.getCanonicalName());
        reset();
    }
    public DTNHost(List<MessageListener> msgLs, List<MovementListener>
            movLs, String groupId, List<NetworkInterface> interfaces,
                   ModuleCommunicationBus comBus, MovementModel mmProto,
                   MessageRouter mRouterProto, boolean hasFileCapability) {
        this.comBus = comBus;
        this.location = new Coord(0, 0);
        this.address = getNextAddress();
        this.name = groupId + address;
        this.net = new ArrayList<NetworkInterface>();
        this.hasFileCapability = hasFileCapability;
        this.allConnections = new ArrayList<Connection>();

        for (NetworkInterface i : interfaces) {
            NetworkInterface ni = i.replicate();
            ni.setHost(this);
            net.add(ni);
        }
        this.msgListeners = msgLs;
        this.movListeners = movLs;
        this.movement = mmProto.replicate();
        this.movement.setComBus(comBus);
        this.movement.setHost(this);
        setRouter(mRouterProto.replicate());
        this.location = movement.getInitialLocation();
        this.nextTimeToMove = movement.nextPathAvailable();
        this.path = null;
        if (movLs != null) {
            for (MovementListener l : movLs) {
                l.initialLocation(this, this.location);
            }
        }
    }
    private synchronized static int getNextAddress() {
        return nextAddress++;
    }
    public static void reset() {
        nextAddress = 0;
    }
    public boolean isMovementActive() {
        return this.movement.isActive();
    }
    public boolean isRadioActive() {
        // Radio is active if any of the network interfaces are active.
        for (final NetworkInterface i : this.net) {
            if (i.isActive())
                return true;
        }
        return false;
    }
    private void setRouter(MessageRouter router) {
        router.init(this, msgListeners);
        this.router = router;
    }
    public MessageRouter getRouter() {
        return this.router;
    }
    @SuppressWarnings("unused")
    private void setFileSystem(DTNFileSystem fileSystem) {
        fileSystem.init(this);
        this.fileSystem = fileSystem;
    }
    public DTNFileSystem getFileSystem() {
        return this.fileSystem;
    }
    public int getAddress() {
        return this.address;
    }
    public ModuleCommunicationBus getComBus() {
        return this.comBus;
    }
    public void connectionUp(Connection con) {
        this.allConnections.add(con);
        this.router.changedConnection(con);
    }
    public void connectionDown(Connection con) {
        this.allConnections.remove(con);
        this.router.changedConnection(con);
    }
    public List<Connection> getConnections() {
        return Collections.unmodifiableList(allConnections);
    }
    public Coord getLocation() {
        return this.location;
    }
    public Path getPath() {
        return this.path;
    }
    public void setLocation(Coord location) {
        this.location = location.clone();
    }
    public void setName(String name) {
        this.name = name;
    }
    public Collection<Message> getMessageCollection() {
        return this.router.getMessageCollection();
    }
    public int getNrofMessages() {
        return this.router.getNrofMessages();
    }
    public double getBufferOccupancy() {
        long bSize = router.getBufferSize();
        long freeBuffer = router.getFreeBufferSize();
        return 100 * ((bSize - freeBuffer) / (bSize * 1.0));
    }
    public RoutingInfo getRoutingInfo() {
        return this.router.getRoutingInfo();
    }
    public List<NetworkInterface> getInterfaces() {
        return net;
    }
    public NetworkInterface getInterface(int interfaceNo) {
        NetworkInterface ni = null;
        try {
            ni = net.get(interfaceNo - 1);
        } catch (IndexOutOfBoundsException ex) {
            throw new SimError(
                    "No such interface: " + interfaceNo + " at " + this);
        }
        return ni;
    }
    protected NetworkInterface getInterface(String interfacetype) {
        for (NetworkInterface ni : net) {
            if (ni.getInterfaceType().equals(interfacetype)) {
                return ni;
            }
        }
	}
    public void forceConnection(DTNHost anotherHost, String interfaceId,
                                boolean up) {
        NetworkInterface ni;
        NetworkInterface no;

        if (interfaceId != null) {
            ni = getInterface(interfaceId);
            no = anotherHost.getInterface(interfaceId);

            assert (ni != null) :
                    "Tried to use a nonexisting interfacetype " + "" + "" +
                            "" + "" + "" + "" + "" + "" + "" + "" + "" + "" +
                            "" + "" + "" + "" + "" + "" + "" + "" + "" + "" +
                            "" + "" + "" + "" + "" + "" + "" + "" + "" +
                            interfaceId;
            assert (no != null) :
                    "Tried to use a nonexisting interfacetype " + "" + "" +
                            "" + "" + "" + "" + "" + "" + "" + "" + "" + "" +
                            "" + "" + "" + "" + "" + "" + "" + "" + "" + "" +
                            "" + "" + "" + "" + "" + "" + "" + "" + "" +
                            interfaceId;
        } else {
            ni = getInterface(1);
            no = anotherHost.getInterface(1);

            assert (ni.getInterfaceType().equals(no.getInterfaceType())) :
                    "Interface types do not match." + "  Please specify " +
                            "interface type explicitly";
        }

        if (up) {
            ni.createConnection(no);
        } else {
            ni.destroyConnection(no);
        }
    }
    public void connect(DTNHost h) {
        if (DEBUG)
            Debug.p("WARNING: using deprecated DTNHost.connect" + "(DTNHost)" +
                    "" + "DTNHost(DTNHost,null,true) instead");
        forceConnection(h, null, true);
    }
    public void update(boolean simulateConnections) {
        if (!isRadioActive()) {
            tearDownAllConnections();
            return;
        }

        if (simulateConnections) {
            for (NetworkInterface i : net) {
                i.update();
            }
        }

        this.router.update();
    }
    private void tearDownAllConnections() {
        for (NetworkInterface i : net) {
            List<Connection> conns = i.getConnections();
            if (conns.size() == 0)
                continue;
            List<NetworkInterface> removeList = new
                    ArrayList<NetworkInterface>(conns.size());
            for (Connection con : conns) {
                removeList.add(con.getOtherInterface(i));
            }
            for (NetworkInterface inf : removeList) {
                i.destroyConnection(inf);
            }
        }
    }
    public void move(double timeIncrement) {
        double possibleMovement;
        double distance;
        double dx, dy;
        Coord newPos;
        double movedDistance;
        if (!isMovementActive() || SimClock.getTime() < this.nextTimeToMove) {
            return;
        }
        if (this.destination == null) {
            if (!setNextWaypoint()) {
                return;
            }
        }
        possibleMovement = timeIncrement * speed;
        distance = this.location.distance(this.destination);
        while (possibleMovement >= distance) {
            this.location.setLocation(this.destination);
            possibleMovement -= distance;
            if (!setNextWaypoint()) {
                return;
            }
            distance = this.location.distance(this.destination);
        }
        dx = (possibleMovement / distance) *
                (this.destination.getX() - this.location.getX());
        dy = (possibleMovement / distance) *
                (this.destination.getY() - this.location.getY());
        if (this.router instanceof EnergyCarrier) {
            newPos = this.location.clone();
            newPos.translate(dx, dy);
            movedDistance = this.location.distance(newPos);
            if (((EnergyCarrier) this.router).getEnergyModule() != null &&
                    ((EnergyCarrier) this.router).getEnergyModule()
                            instanceof MobilityEnergyDepletion) {
                ((MobilityEnergyDepletion) ((EnergyCarrier) this.router)
                        .getEnergyModule()).drainViaMovement(movedDistance);
            }
        }
        this.location.translate(dx, dy);
    }
    private boolean setNextWaypoint() {
        if (path == null) {
            path = movement.getPath();
        }

        if (path == null || !path.hasNext()) {
            this.nextTimeToMove = movement.nextPathAvailable();
            this.path = null;
            return false;
        }

        this.destination = path.getNextWaypoint();
        this.speed = path.getSpeed();

        if (this.movListeners != null) {
            for (MovementListener l : this.movListeners) {
                l.newDestination(this, this.destination, this.speed);
            }
        }

        return true;
    }
    public void sendMessage(String id, DTNHost to) {
        this.router.sendMessage(id, to);
    }
    public int receiveMessage(Message m, DTNHost from) {
        return this.router.receiveMessage(m, from);
    }
    public boolean requestDeliverableMessages(Connection con) {
        return this.router.requestDeliverableMessages(con);
    }
    public void messageTransferred(String id, DTNHost from) {
        this.router.messageTransferred(id, from);
    }
    public void messageAborted(String id, DTNHost from, int bytesRemaining) {
        this.router.messageAborted(id, from, bytesRemaining);
    }
    public void createNewMessage(Message m) {
        this.router.createNewMessage(m);
    }
    public void deleteMessage(String id, boolean drop) {
        this.router.deleteMessage(id, drop);
    }
    public String toString() {
        return name;
    }
    public boolean equals(DTNHost otherHost) {
        return this == otherHost;
    }
    public int compareTo(DTNHost h) {
        return this.getAddress() - h.getAddress();
    }
    public boolean hasFileCapability() {
        return hasFileCapability;
    }
    public int getGroupId(int[] groupSizes) {
        int groupId = 0;
        int i = 0;
        while (i < address) {
            i += groupSizes[groupId];
            groupId++;
        }
        return groupId;
    }
    public MovementModel getMovement() {
        return this.movement;
    }
}
